/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.container;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectInputStream;
import java.lang.Class;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Properties;
import java.util.Vector;
import javax.naming.Context;
import javax.naming.NamingException;
import javax.transaction.TransactionManager;

import org.objectweb.jonas.common.ClassLoaderFactory;
import org.objectweb.jonas.common.Trace;
import org.objectweb.jonas.deployment.api.BeanDesc;
import org.objectweb.jonas.deployment.api.DeploymentDesc;
import org.objectweb.jonas.deployment.api.DeploymentDescException;
import org.objectweb.jonas.deployment.api.EntityJdbcCmpDesc;
import org.objectweb.jonas.deployment.api.SessionDesc;
import org.objectweb.jonas.jtm.TransactionService;
import org.objectweb.jonas.lib.BeanNaming;
import org.objectweb.jonas.naming.CompNamingContext;
import org.objectweb.jonas.naming.NamingManager;


/**
 * This class represents an EJB container as specified in EJB specification 8.2
 *
 * A container is where an enterprise Bean object lives. 
 * Multiple EJB classes can be installed in a single container.
 * For each EJB class installed, the container provides a home interface and
 *  makes this interface available in the JNDI name space.
 */
public class Container {

    static private final int trace = Trace.DB_14;

    static private Hashtable listOfEJBHome     = new Hashtable();
    static private Hashtable listOfDataSources = new Hashtable();
    static TransactionManager current = null;

    /**
     * constructor
     */
    public Container() {

	Trace.outln(trace, "Container.Container() ");
	current = TransactionService.getCurrent();
    }

    /**
     * Adds beans in container.
     *
     * @param desc  DD. 
     *
     */
    public void addBeans(BeanDesc  desc) {
	Trace.outln(trace, "Container.addBeans() ");
	
	boolean   isSession = (desc instanceof SessionDesc);
		
	Trace.outln (trace, "Bean Deployment Descriptor of '" + desc.getEjbName() +
			     "' is read");
	Trace.outln (trace, desc.toString());
		
	// Build the JOnAS generated names for that bean.
	BeanNaming beanNames = new BeanNaming(desc);
		
	// Finds the DataSource associated to the bean
	// Useful for Entity with Container Managed Persistence.
	String dataSourceName = null;
	if (desc instanceof EntityJdbcCmpDesc) {
	    dataSourceName = ((EntityJdbcCmpDesc)desc).getDatasourceJndiName();
	}
	Object dataSource = null;

	Context context = NamingManager.getInitialContext();
	if (dataSourceName != null) {
	    try {
		dataSource = context.lookup(dataSourceName);
	    }catch(Exception  ex){
		Trace.errln("JOnAS Server configuration error!");
		Trace.errln("     Datasource '"+dataSourceName+"' is not known by the EJB server.");
		Trace.errln("     Please check the 'jonas.datasources' property in the 'jonas.properties' file.");  
	    }
	}
	// Loads the specific EJBHome class
	String ejbHomeClassName = beanNames.getFullWrpHomeName();
	Class ejbHomeClass = null;
	try {
	    // Muly changes partly undone
	    //ejbHomeClass = ClassLoaderFactory.getClassLoader().loadClass(ejbHomeClassName);
	    ejbHomeClass = desc.getHomeClass().getClassLoader().loadClass(ejbHomeClassName);
	} catch (ClassNotFoundException e) {
	    Trace.errln("    Cannot load "+ ejbHomeClassName );
	    System.exit(1);
	}
	Trace.outln (trace, ejbHomeClassName+ " loaded");
		
		
	// Creates a unique EJBHome instance by calling the appropriate constructor
	Object ejbHomeInstance  = null;
	String beanName         = desc.getEjbName();
	try {
	    int nbParams = isSession ? 2 : 3;
	    Class paramTypes[]   = new Class[nbParams];  // constructor argument types
	    Object paramObjects[] = new Object[nbParams]; // constructor argument values
		    
	    if (isSession) {
		paramTypes[0]       = org.objectweb.jonas.deployment.api.SessionDesc.class;
	    } else {
		paramTypes[0]       = org.objectweb.jonas.deployment.api.EntityDesc.class;
	    }
	    paramObjects[0]     = (Object)desc;
		    
	    // Second parameter
	    paramTypes[1]       = java.lang.String.class;
	    paramObjects[1]     = (Object)beanName;
		    
	    // Third parameter
	    if (!isSession) {
		paramTypes[2]   = java.lang.Object.class;
		paramObjects[2] = (Object)dataSource;
	    }
		    
	    Constructor constructor = ejbHomeClass.getConstructor(paramTypes);
	    ejbHomeInstance = constructor.newInstance(paramObjects);
		    
	} catch (InvocationTargetException te) {
	    Trace.errln("    InvocationTargetException : " + te.getTargetException());
	    te.getTargetException().printStackTrace();
	} catch (Exception e) {
	    Trace.errln("    Call of " + ejbHomeClassName + " constructor : " + e);
	    e.printStackTrace();
	}

	// Keep the new EJBHome in a list
	listOfEJBHome.put(ejbHomeClassName, ejbHomeInstance);
	Trace.outln (Trace.DB_31, "EJBHome:"+ejbHomeClassName+" for " + beanName + " available");

	// Bean Environment
	try {
	    // Create a JNDI context for this bean home
	    CompNamingContext ctx = new CompNamingContext(beanName);
	    ((JBeanHome)ejbHomeInstance).setJNDIContext(ctx);

	    // setup the bean environment in this context
	    NamingManager.setBeanEnvironment(ctx, desc);
	} catch (NamingException e) {
	    Trace.errln("addBeans: could not setup environment");
	}
    }



    /**
     * Gets Names of the beans deployed in the container.
     *
     * @return A list of beans names
     *
     */
    public Vector listBeans() {
	Trace.outln(trace, "Vector Container.listBeans() ");
	Vector v = new Vector();
	for (Enumeration lk = listOfEJBHome.keys(); lk.hasMoreElements() ;) {
	    v.addElement(lk.nextElement());
	}
	/* debugging
	Enumeration liste = listOfEJBHome.elements();
	while (liste.hasMoreElements()) {
	    JBeanHome h = (JBeanHome) liste.nextElement();
	    Properties beanEnv = h.getEjb10Environment();
	    Trace.outln(trace, "Bean name = "+beanEnv.getProperty("mybeanname", "NO NAME"));
	}
	*/
	return v;
    }

    /**
     * Remove the beans
     *
     */
    public void removeBeans(Vector descNames) {
	Trace.outln(trace, "Container.removeBeans()");
        for (Enumeration lk = descNames.elements(); lk.hasMoreElements();) {
            Object key = lk.nextElement();
            JBeanHome bbh = (JBeanHome) listOfEJBHome.get(key);
            bbh.unregister();
            listOfEJBHome.remove(key);
            bbh = null;
        }
    }


    /**
     * Unbinds all the EJBHome names from JNDI
     *
     */
    public void unbindBeans() {
	Trace.outln(trace, "void Container.unbindBeans()");
	try {
	    for (Enumeration lk = listOfEJBHome.elements(); lk.hasMoreElements() ;) {
		JBeanHome bbh = (JBeanHome)lk.nextElement();
		bbh.unregister();
	    }
	} catch (Exception e) {
	    Trace.errln("void Container.unbindBeans() : " + e);
	    return;
	}
    }


}
