/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999,2000 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * 00/01/08 Markus Fritz
 *          make connection pooling thread safe (with same tx)
 *          introduce MaxAge for connections
 *          optimize pooling
 *          add some 'paranoia code': test conns before returning
 *	    them from the pool.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.dbm;

import java.io.Serializable;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.Vector;
import java.util.Iterator;
import java.util.Enumeration;
import javax.naming.Context;
import javax.naming.NamingException;
import javax.naming.Reference;
import javax.naming.Referenceable;
import javax.naming.StringRefAddr;
import javax.sql.ConnectionEvent;
import javax.sql.ConnectionEventListener;
import javax.sql.DataSource;
import javax.sql.XAConnection;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.xa.XAResource;
import org.objectweb.jonas.common.Trace;
import org.objectweb.jonas.jdbc_xa.XADataSourceImpl;
import org.objectweb.jonas.jtm.CompletedTransactionListener;
import org.objectweb.jonas.jtm.Current;
import org.objectweb.jonas.jtm.TransactionImpl;

/**
 * This class represents a connection stored in the pool.
 * It may be in different states:
 * - available and reuseable for this user (open=0, tx=null)
 * - reusable for the same transaction, not already committed (open=?, tx!=null)
 * - still open, but not involved in a Tx (open=1, tx=null)
 * - busy (open>jdbcConnMaxThreads, tx!=null)
 * These 2 cases must be considered: close after commit, commit after close.
 * We cannot suppose what shema the user will use and we must handle both correctly.
 */
class PoolItem {
    public XAConnection xaConn; // the pooled connection
    public String user;         // the user name
    public int open;            // >0 if connection is open
    public Transaction tx;  // transaction the connection is involved with
    public long lMaxAgeMillis;  // When it is expected to die
    
    /*
     * constructor used by newConnection()
     */
    public PoolItem(XAConnection xac, String user, long maxage) {
        this.xaConn = xac;
        this.user = user;
        open = 0;
        lMaxAgeMillis = System.currentTimeMillis() + maxage;
    }
}

class Pool {
    private static final int trace = Trace.DB_8;
    
    // Liste of free available items
    private LinkedList oFreeItems = new LinkedList();
    // requests with no tx are identified by thread
    private int iItems_w_o_Tx = 0;
    // Requests with same tx get always the same conn
    private Hashtable hTx2Item = new Hashtable();
    
    // map XAConn => Item (to speedup CloseConn)
    private Hashtable hXAC2Item = new Hashtable();

    private ConnectionManager myConnMgr = null;
    
    // ---------------------------------------------------------------------
    // Pool Configuration
    // ---------------------------------------------------------------------
    // Max age of a Connection in milliseconds.
    long jdbcConnMaxAge;
    // Max. concurrent threads on same tx/conn
    int jdbcConnMaxThreads;
    // Max wait time (in milliseconds) if more than jdbcConnMaxThreads threads request conn
    long jdbcConnExclTimeout;
    // Level of checking on connections
    int jdbcConnCheckLevel;
    // test statement
    String jdbcTestStmt;

    public Pool(ConnectionManager cmgr) {
	myConnMgr = cmgr;
	(new PoolKeeper(this)).start();
    }
    
    /**
     * Look up a free XAConnection for this user and this transaction.
     * If found, set the item to "busy".
     *
     * @return the free PoolItem or null if no item available
     */
    public PoolItem openConnection(String user, Transaction  tx) throws SQLException {
        PoolItem item = null;
	boolean newTx = false;
        synchronized(this) {
            if (tx != null) {
                Object i;
                while ((i = hTx2Item.get(tx)) != null) {
                    if (i instanceof PoolItem)
                        break;
                    else try {
                        wait(); // for placeholder to be replaced
                    } catch (InterruptedException ign) {
                    }
		}
                item = (PoolItem)i;
            }
            if (item == null) {
		while (true) {
		    synchronized (oFreeItems) {
			if (oFreeItems.isEmpty()) {
			    Trace.outln(trace, 
					 "  Pool: cannot find a reusable Connection for Tx = " 
					 + String.valueOf(tx));
			    if (tx != null)
				hTx2Item.put(tx, new Object()); // Placeholder
			    return null; // => newConnection
			}
			item = (PoolItem)oFreeItems.removeFirst();
		    }
		    if (jdbcConnCheckLevel > 0) {
			try {
			    Connection conn = item.xaConn.getConnection();
			    if (conn.isClosed()) { 
				destroyItem(item); 
				continue; 
			    }
			    if (jdbcConnCheckLevel > 1) {
				Trace.outln(trace, "Retrying connection: "+jdbcTestStmt);
				java.sql.Statement stmt = conn.createStatement();
				stmt.execute(jdbcTestStmt); // => exception
				stmt.close();
			    }
			} catch (Exception ex) {
			    Trace.errln("  Pool: Remove invalid item.\n\t"+ex);
			    destroyItem(item); 
			    continue;
			}
		    }
		    break;
		}
		item.tx = tx;

                if (tx == null) {
                    iItems_w_o_Tx++;
                } else {
                    hTx2Item.put(tx, item);
		    newTx = true;
		}
            }
        }
        
        if (tx != null) {
            Trace.outln(trace, "  Pool: reuse a Connection for Tx = "+ String.valueOf(tx));
	    if (newTx) {
		try {
		    ((TransactionImpl)tx).addCompletedTransactionListener(myConnMgr);
		} catch (SystemException e) {
		    Trace.errln("openConnection: cannot addCompletedTransactionListener:"+e);
		}
	    }
	}

        // Try to get Connection open max jdbcConnMaxThreads times concurrent
        long w, t = System.currentTimeMillis() + jdbcConnExclTimeout;
        synchronized (item) {
            // Handle State "busy" (open>jdbcConnMaxThreads, tx!=null) => wait
            while (item.open >= jdbcConnMaxThreads) {
                w = t - System.currentTimeMillis();
                if (w < 0) {
                    Trace.errln("WARNING: Cannot 'exclusively' get connection");
                    break; // return non-exclusive conn
                }
                
                Trace.outln(trace, "  Pool: waiting for exclusive connection");
                try {
                    item.wait(w);
                } catch (InterruptedException ign) {
                    Trace.errln("WARNING: Interupted while waiting for"+
                                 " exclusive connection");
                    break; // return non-exclusive conn
                }
            }
            item.open++;
        }
        return item;
    }
    
    /**
     * Add a new XAConnection in the pool
     */

    public PoolItem  newConnection(XAConnection xac, String user, Transaction tx) {
        PoolItem item = new PoolItem(xac, user, jdbcConnMaxAge);
        hXAC2Item.put(xac, item);
        item.tx = tx;
        item.open++;
        synchronized(this) {
            if (tx == null) {
                iItems_w_o_Tx++;
            } else {
                hTx2Item.remove(tx); // placeholder
                hTx2Item.put(tx, item);
                notifyAll();
            }
        }
        Trace.outln(trace, "  Pool: add a Connection for Tx = " + String.valueOf(tx));

	if (tx != null) {
	    try {
		((TransactionImpl)tx).addCompletedTransactionListener(myConnMgr);
	    } catch (SystemException e) {
		Trace.errln("newConnection: cannot addCompletedTransactionListener:"+e);
	    }
	}
        return item;
    }
    
    /**
     * The transaction has committed (or rolled back). We can return its
     * connections to the pool of available connections.
     */
    public void freeConnections(Transaction tx) {

        // tx is guarantied to be not null
        PoolItem item = (PoolItem)hTx2Item.remove(tx);
        if (item == null) {
            Trace.errln("  Pool: no connection found to free for this Tx");
	    return;
        }

	synchronized(item) {
            Trace.outln(trace, "  Pool: free a Connection for this Tx");
            item.tx = null;

	    // This code must be changed (it's not correct!)
            if (item.open > 0)
                Trace.errln("WARNING: Connection was not always closed correctly !");
            item.open = 0;
            
            // Check if this conn has reached it's max age
            if (item.lMaxAgeMillis < System.currentTimeMillis()) {
                Trace.outln(trace, 
                             "  Pool: remove a Connection from this pool"+
                             " while its max age is reached");
		destroyItem(item);
            } else synchronized (oFreeItems) {
                // Add it to Free list
                oFreeItems.addLast(item);
            }
        }
    }
    

    /**
     * Mark a specific Connection in the pool as closed.
     * If it is no longer associated to a Tx, we can free it.
     *
     * @return the PoolItem the connection is in (=> tx)
     */
    public PoolItem closeConnection(XAConnection xac) {
        PoolItem item = (PoolItem)hXAC2Item.get(xac);
        if (item == null) {
            Trace.errln("Cannot resolve XAC => PoolItem");
            return null;
        }
        
        // The connection will be available only if not associated
        // to a transaction. Else, it will be reusable only for the
        // same transaction.
        synchronized(item) {
            item.open--;
            if (item.open < 0) {
                item.open = 0;
                Trace.errln("connection-open counter underflow");
            } else if (item.tx == null && item.open > 0) {
                item.open = 0;
                Trace.errln("connection-open counter overflow");
            }

            if (item.tx != null) {
                Trace.outln(trace, "  Pool: keep the Connection for this tx");
            } else {
                Trace.outln(trace, "  Pool: free the Connection");

                // while freeConnections is not called, do it here
                iItems_w_o_Tx--;

                // Check if this conn has reached its max age
                if (item.lMaxAgeMillis < System.currentTimeMillis()) {
                    Trace.outln(trace, 
                                 "  Pool: remove a Connection from this pool"+
                                 " while it's max age is reached");
		    destroyItem(item);
                } else synchronized (oFreeItems) {
                    // else add it to FreeList for reuse
                    oFreeItems.addLast(item);
                }
            }
            
            item.notify();
        }
        return item;
    }

    private void destroyItem(PoolItem item) {
       hXAC2Item.remove(item.xaConn);
       try {
           item.xaConn.close();
       } catch (java.sql.SQLException ign) {
           Trace.errln(ign.toString());
       }
       item.xaConn = null;
    }

    /**
     * Check all connections not in a transaction if their maxage
     * is reached and if so kill them.
     */
    public void check4MaxAge() {
        PoolItem item;
        synchronized (oFreeItems) {
            Iterator i = oFreeItems.iterator();
            while (i.hasNext()) {
                item = (PoolItem) i.next();
                synchronized(item) {
                    if (item.tx != null || item.open > 0)
                        Trace.errln("Inconsistent PoolItem removed from FreeList");
                    // Ceck if this conn has reached it's max age
                    if (item.lMaxAgeMillis < System.currentTimeMillis()) {
                        Trace.outln(trace, 
                                     "  Pool: remove a Connection from this pool"+
                                     " while it's max age is reached");
                        i.remove();
			destroyItem(item);
                    }
                }
            }
        }
        // For debugging: print statistics
        if (false) synchronized (this) {
            if (hXAC2Item.size() 
                 != oFreeItems.size() + iItems_w_o_Tx + hTx2Item.size()) {
                Trace.errln("  Pool: Inconsistency detected !");
                
                Enumeration e = hXAC2Item.keys();
                Object k;
                Trace.errln("\tPopulation:");
                while (e.hasMoreElements()) {
                    k = e.nextElement();
                    item = (PoolItem)hXAC2Item.get(k);
                    Trace.errln("\t\t" + String.valueOf(k) + "\t" + 
                                 String.valueOf(item));
                }
                
                Iterator i = oFreeItems.iterator();
                Trace.errln("\tFreeList:");
                while (i.hasNext()) {
                    item = (PoolItem)i.next();
                    Trace.errln("\t\t" + String.valueOf(item));
                }
                
                Trace.errln("\tw/o Tx:\t\t"+String.valueOf(iItems_w_o_Tx));
                
                e = hTx2Item.keys();
                Trace.errln("\twith Tx:");
                while (e.hasMoreElements()) {
                    k = e.nextElement();
                    item = (PoolItem)hTx2Item.get(k);
                    Trace.errln("\t\t" + String.valueOf(k) + "\t" + 
                                 String.valueOf(item));
                }
                
                // for debugging:
                System.exit(1);
            } else
                Trace.errln("  Pool: Statistics\n\tPopulation:\t"+
                             String.valueOf(hXAC2Item.size()) + 
                             "\n\tFree:\t\t" +
                             String.valueOf(oFreeItems.size()) + 
                             "\n\tw/o Tx:\t\t" +
                             String.valueOf(iItems_w_o_Tx) +
                             "\n\twith Tx:\t" +
                             String.valueOf(hTx2Item.size()));
        }
    }
    
}
    
/**
 * Keeps pool as small as possible (after conn's has reached MaxAge)
 */
class PoolKeeper extends Thread {

    private static final int trace = Trace.DB_8;

    protected Pool pool;

    public PoolKeeper(Pool pool) {
       super("PoolKeeper");
       setDaemon(true);
       this.pool = pool;
    }

    public void run() {

       Trace.outln(trace, "  PoolKeeper: started");

       while (true) {
	   try {
	       sleep(60 * 1000); // 60 sec
	       pool.check4MaxAge();
	   } catch (Exception ex) {
	       Trace.errln("  PoolKeeper: " + ex.toString());
	   }
       }
    }
}

/**
 * JOnAS generic implementation of DataSource to manage an underlying
 * JDBC-XA driver (XADataSource, XAConnection, XAResource, ...)
 * This class acts as a factory for Connection objects.
 * Its main goal is to manage a pool of XAConnection objects, to avoid 
 * closing the actual Connection on DBMS, and to deal with distributed
 * transactions.
 * This DataSource is registered in JNDI (=> implements Referenceable)
 */
public class ConnectionManager implements DataSource, 
					  Referenceable, 
					  Serializable,
					  CompletedTransactionListener,
                                          ConnectionEventListener {

    private static final int trace = Trace.DB_8;

    // The static table of all the ConnectionManager objects
    // Each ResourceManager has its own ConnectionManager (i.e. DataSource)
    private static Hashtable cmList = new Hashtable();

    /**
     * unique Current object, kept here for efficiency
     * @serial
     */
    private Current current = Current.getCurrent();

    /**
     * underlaying XA DataSource
     * @serial
     */
    private XADataSourceImpl xads = null;

    /**
     * The pool of XAConnections managed by this Connection Manager
     * @serial
     */
    private Pool pool = new Pool(this);

    // ----------------------------------------------------------
    // properties
    // ----------------------------------------------------------

    /**
     * @serial for JNDI
     */
    private String dataSourceName = null;
    /**
     * @serial url for database
     */
    private String url = null;
    /**
     * @serial JDBC driver Class
     */
    private String className = null;
    /**
     * @serial default user
     */
    private String userName = null;
    /**
     * @serial default passwd
     */
    private String password = null;

    // -----------------------------------------------------------------
    // Constructors
    // -----------------------------------------------------------------

    /**
     * Constructor for Factory
     */
    public ConnectionManager() {
        Trace.outln(trace, "ConnectionManager.ConnectionManager()");

        // Create the underlaying JDBC-XA driver
        // We do not use JNDI here. (should we do ?)
        xads = new XADataSourceImpl();
    }

    // -----------------------------------------------------------------
    // Properties
    // -----------------------------------------------------------------
    public String getDataSourceName() {
        return dataSourceName;
    }
    public void setDataSourceName(String s) {
        Trace.outln(trace, "ConnectionManager.setDataSourceName: "+s);
        dataSourceName = s;
        xads.setDataSourceName(s); // for getRMID!

        // Add it to the list
        cmList.put(s, this);
    }

    public String getUrl() {
        return url;
    }
    public void setUrl(String s) {
        Trace.outln(trace, "ConnectionManager.setUrl(String "+s+")");
        url = s;
        // Forward it to JDBC-XA driver
        xads.setUrl(s);
    }

    public String getClassName() {
        return className;
    }
    public void setClassName(String s) {
        Trace.outln(trace, "ConnectionManager.setClassName(String "+s+")");
        className = s;
        // Forward it to JDBC-XA driver
        xads.setClassName(s);
    }

    public String getUserName() {
        return userName;
    }
    public void setUserName(String s) {
        Trace.outln(trace, "ConnectionManager.setUserName(String "+s+")");
        userName = s;
        // Forward it to JDBC-XA driver
        xads.setUserName(s);
    }

    public String getPassword() {
        return password;
    }
    public void setPassword(String s) {
        Trace.outln(trace, "ConnectionManager.setPassword(String "+s+")");
        password = s;
        // Forward it to JDBC-XA driver
        xads.setPassword(s);
    }

    /**
     * Configure the Connection pool. Called by the Container at init.
     * Configuration can be set in datasource.properties files.
     */
    public void poolConfigure(String connchecklevel,
			      String connmaxage,
			      String connmaxthreads,
			      String connexcltimeout,
			      String connteststmt) {
	pool.jdbcConnCheckLevel = (new Integer(connchecklevel)).intValue();
	pool.jdbcConnMaxAge = (new Integer(connmaxage)).intValue();
	pool.jdbcConnMaxThreads = (new Integer(connmaxthreads)).intValue();
	pool.jdbcConnExclTimeout = (new Integer(connexcltimeout)).intValue();
	pool.jdbcTestStmt = connteststmt;
	Trace.outln(trace, "ConnectionManager configured with:");
	Trace.outln(trace, "   jdbcConnCheckLevel  = "+pool.jdbcConnCheckLevel);
	Trace.outln(trace, "   jdbcConnMaxAge      = "+pool.jdbcConnMaxAge);
	Trace.outln(trace, "   jdbcConnMaxThreads  = "+pool.jdbcConnMaxThreads);
	Trace.outln(trace, "   jdbcConnExclTimeout = "+pool.jdbcConnExclTimeout);
	Trace.outln(trace, "   jdbcTestStmt        = "+pool.jdbcTestStmt);
	// set times in milliseconds
	pool.jdbcConnMaxAge = pool.jdbcConnMaxAge * 60L * 1000L;
	pool.jdbcConnExclTimeout = pool.jdbcConnExclTimeout * 1000L;
    }

    // -----------------------------------------------------------------
    // DataSource implementation
    // -----------------------------------------------------------------

    /**
     * Attempt to establish a database connection.
     *
     * @return a Connection to the database
     *
     * @exception java.sql.SQLException - if a database-access error occurs.
     */
    public Connection getConnection() throws SQLException {
        Trace.outln(trace, "ConnectionManager.getConnection()");
        return getConnection(userName, password);
    }

    /**
     * Attempt to establish a database connection.
     * The pool management must be "synchronized".
     *
     * @param user - the database user on whose behalf the Connection is being made
     * @param password - the user's password
     *
     * @return a Connection to the database
     *
     * @exception SQLException - if a database-access error occurs.
     */
    public synchronized Connection getConnection(String username, String password) 
        throws SQLException {

        Trace.outln(trace, "ConnectionManager.getConnection("+username+")");

        XAConnection xac = null;
        Transaction tx = null;
        try {
            tx = current.getTransaction();
            Trace.outln(trace, "\tTx = " + String.valueOf(tx));
	} catch (NullPointerException n) {
	    // current is null: we are not in EJBServer.
	    Trace.errln("ConnectionManager: should not be used outside an EJBServer");
        } catch (SystemException e) {
            Trace.errln("ConnectionManager: getTransaction failed:"+e);
        }

        // Check if available XAConnection in the pool for this user

        PoolItem pi = pool.openConnection(username, (Transaction) tx);
        if (pi == null) {
            Trace.outln(trace, "ConnectionManager: No more XAConnection in the pool");
            // Get a new XAConnection
            xac = xads.getXAConnection(username, password);
            // Put it in the pool as in-use

	    pi = pool.newConnection(xac, username, (Transaction) tx);
            // Register me as a ConnectionEventListener
            xac.addConnectionEventListener(this);
        } else {
            // We found a reusable XAConnection
            // We are still registered as ConnectionEventListener
            Trace.outln(trace, "ConnectionManager: Reuse an XAConnection from the pool");
            xac = pi.xaConn;
        }

        synchronized (xac) {
	    Connection ret = xac.getConnection();
	    // Enlist XAResource if we are actually in a transaction
	    if (tx != null) {
		if (pi.open == 1) { // Only if first/only thread
		    // We are inside a transaction: no autocommit!
		    ret.setAutoCommit(false);
		    try {
			tx.enlistResource(xac.getXAResource());
		    } catch (Exception e) {
			Trace.errln("Cannot enlist XAResource:"+e);
			// What shall we do here ? XXX
		    }
		}
	    } else {
		// We are not in a transaction yet: we just make a pre-registration.
		ret.setAutoCommit(true);    // in case we do not start a Tx
		current.addConnection(xac);
	    }

	    // return a Connection object
	    return ret;
	}
    }

    /**
     * Get the log writer for this data source. 
     * 
     * The log writer is a character output stream to which all logging and 
     * tracing messages for this data source object instance will be printed. 
     * This includes messages printed by the methods of this object, messages 
     * printed by methods of other objects manufactured by this object, and so on.
     * Messages printed to a data source specificlog writer are not printed to the 
     * log writer associated with the Drivermanager class. When a data
     * source object is created the log writer is initially null, in other words, 
     * logging is disabled.
     *
     * @return the log writer for this data source, null if disabled
     *
     * @exception SQLException - if a database-access error occurs.
     */
    public java.io.PrintWriter getLogWriter() throws SQLException {
        Trace.outln(trace, "ConnectionManager.getLogWriter()");
        return xads.getLogWriter();
    }


    /**
     * Set the log writer for this data source. 
     * 
     * The log writer is a character output stream to which all logging and 
     * tracing messages for this data source object instance will be printed. 
     * This includes messages printed by the methods of this object, messages printed
     * by methods of other objects manufactured by this object, and so on. 
     * Messages printed to a data source specific log writer are not printed to 
     * the log writer associated with the Drivermanager class. When a data
     * source object is created the log writer is initially null, in other words, 
     * logging is disabled.
     * 
     * @param out - the new log writer; to disable, set to null
     *
     * @exception SQLException - if a database-access error occurs.
     */
    public void setLogWriter(java.io.PrintWriter out) throws SQLException {
        Trace.outln(trace, "ConnectionManager.setLogWriter()");
        xads.setLogWriter(out);
    }


    /**
     * Sets the maximum time in seconds that this data source will wait while 
     * attempting to connect to a database.
     * A value of zero specifies that the timeout is the default system timeout 
     * if there is one; 
     * otherwise it specifies that there is no timeout. 
     * When a data source object is created the login timeout is initially zero.
     *
     * @param seconds - the data source login time limit
     *
     * @exception SQLException - if a database access error occurs.
     */
    public void setLoginTimeout(int seconds) throws SQLException {
        Trace.outln(trace, "ConnectionManager.setLoginTimeout(int"+seconds+")");
        xads.setLoginTimeout(seconds);
    }

    /**
     * Gets the maximum time in seconds that this data source can wait while attempting 
     * to connect to a database.
     * A value of zero means that the timeout is the default system timeout if there is one; 
     * otherwise it means that there is no timeout. When a data source object is created 
     * the login timeout is initially zero.
     *
     * @return the data source login time limit
     *
     * @exception SQLException - if a database access error occurs.
     */
    public int getLoginTimeout() throws SQLException {
        Trace.outln(trace, "ConnectionManager.getLoginTimeout()");
        return xads.getLoginTimeout();
    }

    // -----------------------------------------------------------------
    // ConnectionEventListener implementation
    // -----------------------------------------------------------------

    /**
     * Invoked when the application calls close() on its representation 
     * of the connection.
     * @param event - an event object describing the source of the event
     */
    public void connectionClosed(ConnectionEvent event) {
        Trace.outln(trace, "ConnectionManager.connectionClosed(ConnectionEvent)");

        // First of all, always mark the Connection as closed.
        XAConnection xac = (XAConnection) event.getSource();
        synchronized (xac) {
            PoolItem pi = pool.closeConnection(xac);

            Transaction tx = pi == null ? null : (Transaction)pi.tx; 
	    // tx should be the same than getTransaction() ? XXX

	    // delist Resource if in transaction
	    // We must keep the connection till commit or rollback
	    if (tx != null) {
		if (pi.open < 1) try {
                    tx.delistResource(xac.getXAResource(), 
                                       XAResource.TMSUCCESS);
		} catch (Exception e) {
		    Trace.errln("Exception in connectionClosed:"+e);
		}
	    } else {
		current.removeConnection(xac);
	    }
	}
    }


    /**
     * Invoked when a fatal connection error occurs, just before an
     * SQLException is 
     * thrown to the application.
     *
     * @param event - an event object describing the source of the event
     */
    public void connectionErrorOccurred(ConnectionEvent event) {
        Trace.outln(trace, "ConnectionManager.connectionErrorOccurred(ConnectionEvent)");

        // First of all, always mark the Connection as closed.
        XAConnection xac = (XAConnection) event.getSource();

        synchronized (xac) {

            PoolItem pi = pool.closeConnection(xac);

            Transaction tx = pi == null ? null : (Transaction) pi.tx;
            // delist Resource if in transaction
            if (tx != null) {
                if (pi.open < 1) try {
                    tx.delistResource(xac.getXAResource(), XAResource.TMFAIL);
                } catch (Exception e) {
                    Trace.errln("Exception in connectionErrorOccurred:"+e);
                }
            } else {
                // Not sure we must do this...
                current.removeConnection(xac);
            }
        }
    }

    // -----------------------------------------------------------------
    // CompletedTransactionListener interface
    // -----------------------------------------------------------------

    /**
     * Forget all about this Transaction (after the commit or rollback)
     * @param tx the transaction
     * @param st its status
     */
    public void endTransaction(Transaction tx, int st) {

        Trace.outln(trace, "ConnectionManager.endTransaction");

        pool.freeConnections(tx);
    }

    // -----------------------------------------------------------------
    // Referenceable implementation
    // -----------------------------------------------------------------

    /**
     * Retrieves the Reference of this object. Used at binding time by JNDI
     * to build a reference on this object.
     * 
     * @return  The non-null Reference of this object. 
     * @exception  NamingException  If a naming exception was encountered while 
     * retrieving the reference.  
     */
    public Reference getReference() throws NamingException {

        Reference ref = new Reference (this.getClass().getName(), 
                                       "org.objectweb.jonas.dbm.DataSourceFactory",
                                       null);
        // Actually, only the datasource.name is used! (See DataSourceFactory.java)
        ref.add(new StringRefAddr("datasource.name", getDataSourceName()));
        ref.add(new StringRefAddr("datasource.url", getUrl()));
        ref.add(new StringRefAddr("datasource.classname", getClassName()));
        ref.add(new StringRefAddr("datasource.username", getUserName()));
        ref.add(new StringRefAddr("datasource.password", getPassword()));
        return ref;
    }

   /**
    * get the ConnectionManager matching the DataSource name
    */
   public static ConnectionManager getConnectionManager(String dsname) {
       
       ConnectionManager cm = (ConnectionManager) cmList.get(dsname);
       return cm;
   }

}
