/*
 *
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.naming;

import java.util.Hashtable;
import java.util.Properties;
import javax.naming.InitialContext;
import javax.naming.LinkRef;
import javax.naming.NamingException;
import javax.naming.RefAddr;
import javax.sql.DataSource;
import org.objectweb.jonas.common.Trace;
import org.objectweb.jonas.deployment.api.BeanDesc;
import org.objectweb.jonas.deployment.api.EjbRefDesc;
import org.objectweb.jonas.deployment.api.EnvEntryDesc;
import org.objectweb.jonas.deployment.api.ResourceRefDesc;

/**
 * Naming Manager for an EJB Server.
 * this singleton class must exist in each jonas server.
 * all methods are static
 */
public class NamingManager {

    private static final int trace = Trace.DB_3;
    private static final String JNDI_PREFIX = "java:comp/env/";

    // Naming Context associated with the thread
    private static ThreadLocal threadContext = new ThreadLocal();

    // Contexts
    private static InitialContext ictx = null;
    private static CompNamingContext serverContext = null;

    // Env.
    private static Hashtable myEnv = null;

    /**
     * Create the naming manager.
     * useful to create InitialContext because it may throw an exception.
     */
    public NamingManager() throws NamingException {

	// Create the initial context
	// Don't use jonas.properties anymore for jndi properties
	// jndi.properties will be used instead.
	try {
	    ictx = new InitialContext();
	    myEnv = ictx.getEnvironment();
	} catch (NamingException n) {
	    Throwable t = n.getRootCause();
	    if (t.getMessage().startsWith("Connection refused to host:")) {
		Trace.errln("NamingManager: rmi registry not started ?");
		throw n;
	    } else if (t.getMessage().startsWith("error during remote invocation")) {
		Trace.errln("NamingManager: jrmi registry not started ?");
		throw n;
	    } else {
		Trace.errln("NamingManager: "+t.getMessage());
		throw n;
	    }
	}

	// Create a CompNamingContext global for this server
	// not really used today, but could be useful later.
	CompNamingContext serverContext = new CompNamingContext("server", myEnv);
    }

    /**
     * Return the environment for JNDI
     */
    static public Hashtable getEnv() {
	return myEnv;
    }

    /**
     * return the initialContext used in this jonas server
     */
    static public InitialContext getInitialContext() {
	return ictx;
    }

    /**
     * gets the server component context
     */
    static public CompNamingContext getServerContext() {
	return serverContext;
    }

    /*
     * get the DataSource given its name in JNDI
     */
    static public DataSource getDataSource(String db) {

	// lookup the DataSource in the initial context
	DataSource ds = null;
	try {
	    ds = (DataSource) ictx.lookup(db);
	} catch (NamingException e) {
	    Trace.errln("Cannot lookup datasource "+db+": "+e);
	}

	// return it
	return ds;
    }

    /**
     * return the CompNamingContext associated with the current thread.
     */
    static public CompNamingContext getComponentContext() {

	CompNamingContext ctx = (CompNamingContext) threadContext.get();
	if (ctx == null) {
	    Trace.outln(trace, "NamingManager: getComponentContext: null context");
	} else {
	    Trace.outln(trace, "NamingManager: getComponentContext: "+ctx.getCompId());
	}
	return ctx;
    }

    /**
     * associate this CompNamingContext with the current thread.
     * this method should be called in preinvoke/postinvoke
     * and when we build the bean environment.
     */
    static public CompNamingContext setComponentContext(CompNamingContext ctx) {
	if (ctx == null) {
	    Trace.outln(trace, "NamingManager: setComponentContext: reset to null");
	} else {
	    Trace.outln(trace, "NamingManager: setComponentContext: "+ctx.getCompId());
	}
	CompNamingContext ret = (CompNamingContext) threadContext.get();
	threadContext.set(ctx);
	return ret;
    }

    /**
     * Set the bean environment
     * args ctx CompNamingContext for this bean
     * args dd Bean Deployment Descriptor
     * throws NamingException if could not rebind objects
     */
    static public void setBeanEnvironment(CompNamingContext ctx, BeanDesc dd) throws NamingException {

	Trace.outln(trace,"NamingManager: setBeanEnvironment");

	// Set bean context
	CompNamingContext ctxold = setComponentContext(ctx);

	// Bean Environment
	EnvEntryDesc [] envt = dd.getEnvEntryDesc();
	for (int i = 0; i < envt.length; i++) {
	    // get information in descriptor
	    String name = envt[i].getName();
	    Object obj = envt[i].getValue();
	    // register object in JNDI
	    Trace.outln(trace, "NamingManager: Binding object "+name+" -> "+obj);
	    ctx.rebind(name, obj);
	}

	// Resource References
	ResourceRefDesc [] resref = dd.getResourceRefDesc();
	for (int i = 0; i < resref.length; i++) {
	    // get information in descriptor
	    String name = resref[i].getName();
	    String resname = resref[i].getJndiName();
	    // build the LinkRef that will be registered:
	    // FactoryClassName = null, size = 1, refAddr = resname.
	    LinkRef lref = new LinkRef(resname);
	    // register object in JNDI
	    Trace.outln(trace, "NamingManager: Linking resource "+name+" -> "+resname);
	    ctx.rebind(name, lref);
	}
	
	// EJB References
	EjbRefDesc [] ejbref = dd.getEjbRefDesc();
	for (int i = 0; i < ejbref.length; i++) {
	    // get information in descriptor
	    String name = ejbref[i].getName();
	    String ejbname = ejbref[i].getJndiName();
	    LinkRef lref = new LinkRef(ejbname);
	    Trace.outln(trace, "NamingManager: Linking ejb "+name+" -> "+ejbname);
	    ctx.rebind(name, lref);
	}

	// Reset bean context
	setComponentContext(ctxold);
    }

    /**
     * lookup an object in initialContext
     */
    static public Object lookup(String name) throws NamingException {
	Trace.outln(trace, "NamingManager: lookup "+name);
	return ictx.lookup(name);
    }

    /**
     * rebind object in initialContext
     */
    static public void rebind(String name, Object obj) throws NamingException {
	Trace.outln(trace, "NamingManager: rebind "+name);
	ictx.rebind(name, obj);
    }

    /**
     * register a bean in initialContext
     */
    static public void registerBean(BeanDesc dd, Object home) throws NamingException {
	String name = dd.getJndiName();
	ictx.rebind(name, home);
	Trace.outln(trace,"NamingManager.registerBean: "+name);
    }

    /**
     * unregister a bean in initialContext
     */
    static public void unregisterBean(BeanDesc dd) throws NamingException {
	String name = dd.getJndiName();
	ictx.unbind(name);
    }

}
